""" Loads association filters."""

from ...selectors.filters.base_filters import thresh_filter
from ...selectors.filters.qualitative_filters import cramerv_filter, tschuprowt_filter
from ...selectors.filters.quantitative_filters import pearson_filter, spearman_filter
