""" Loads association measures."""

from .base_measures import dtype_measure, make_measure, mode_measure, nans_measure
from .qualitative_measures import chi2_measure, cramerv_measure, tschuprowt_measure
from .quantitative_measures import (
    R_measure,
    distance_measure,
    iqr_measure,
    kruskal_measure,
    zscore_measure,
)
